from typing import Union
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobClient, ContainerClient, BlobServiceClient
from .client_interface import ClientInterface


class AzBlobClient(ClientInterface):

    def _get_service_client_from_credential(
            self,
            account_url: str,
            credential: Union[DefaultAzureCredential, str]) -> BlobServiceClient:
        """
        get BlobServiceClient

        Args:
            account_url:
            credential:

        Returns:
            BlobServiceClient
        """
        return BlobServiceClient(account_url=account_url, credential=credential)

    def _get_service_client_from_connection_string(
            self,
            connection_string: str):
        return BlobServiceClient.from_connection_string(conn_str=connection_string)

    def _get_file_client(
            self,
            account_url: str,
            file_system: str,
            file_path: str) -> BlobClient:
        """
        get BlobClient

        Args:
            account_url:
            file_system:
            file_path:

        Returns:
            BlobClient
        """
        file_client = self._get_service_client_from_url(
            account_url=account_url,
        ).get_blob_client(
            container=file_system,
            blob=file_path
        )
        return file_client

    def _get_container_client(
            self,
            account_url: str,
            file_system: str) -> ContainerClient:
        """
        get ContainerClient

        Args:
            account_url:
            file_system:

        Returns:
            ContainerClient
        """
        container_client = self._get_service_client_from_url(
            account_url=account_url,
        ).get_container_client(
            container=file_system
        )
        return container_client

    def _ls(self, path: str, file_path: str):
        blob_list = \
            [f.name for f in self.get_container_client_from_path(path=path).list_blobs(name_starts_with=file_path)]
        return blob_list

    def _get(self, path: str, offset: int = None, length: int = None, **kwargs):
        file_bytes = self.get_file_client_from_path(path=path).download_blob(offset=offset, length=length).readall()
        return file_bytes

    def _put(self, path: str, data):
        self.get_file_client_from_path(path=path).upload_blob(
            data=data,
            length=len(data),
            overwrite=True
        )
        return True

    def _create(self, path: str):
        raise NotImplementedError

    def _append(self, path: str, data, offset: int):
        raise NotImplementedError

    def _info(self, path: str):
        return self.get_file_client_from_path(path=path).get_blob_properties()

    def _rm(self, path: str):
        self.get_file_client_from_path(path=path).delete_blob()
        return True



from azfs.az_file_client import (
    AzFileClient
)

from azfs.az_file_system import AzFileSystem

from azfs.utils import BlobPathDecoder

from .table_storage import (
    TableStorage,
    TableStorageWrapper
)

# comparable tuple
VERSION = (0, 2, 8)
# generate __version__ via VERSION tuple
__version__ = ".".join(map(str, VERSION))

__all__ = ["AzFileClient", "AzFileSystem", "BlobPathDecoder", "TableStorage", "TableStorageWrapper"]



import pytest
from azfs.clients.blob_client import AzBlobClient
from azure.storage.blob import BlobClient, ContainerClient
import azfs

credential = ""
connection_string = "DefaultEndpointsProtocol=https;AccountName=xxxx;AccountKey=xxxx;EndpointSuffix=core.windows.net"
azc = azfs.AzFileClient()
blob_client_credential = AzBlobClient(credential=credential)
blob_client_connection_string = AzBlobClient(credential=None, connection_string=connection_string)
test_file_path = "https://test.blob.core.windows.net/test/test.csv"
test_file_ls_path = "https://test.blob.core.windows.net/test/"


class BlobMock:
    # dummy blob file class
    def __init__(self, name):
        self.name = name


@pytest.mark.parametrize("blob_client", [
    # blob client
    blob_client_credential,
    blob_client_connection_string,
])
def test_blob_info(mocker, blob_client):
    # ===================== #
    # test for AzBlobClient #
    # ===================== #

    # mock
    func_mock = mocker.MagicMock()
    func_mock.return_value = {
            "name": "test.csv",
            "size": 500,
            "creation_time": "creation_time",
            "last_modified": "last_modified",
            "etag": "etag",
            "content_settings": {
                "content_type": "content"
            }
        }

    mocker.patch.object(BlobClient, "get_blob_properties", func_mock)
    info = blob_client.info(path=test_file_path)
    assert "name" in info

    # ===================== #
    # test for AzFileClient #
    # ===================== #
    info = azc.info(path=test_file_path)
    assert "name" in info
    size = azc.size(path=test_file_path)
    assert size == 500
    check_sum = azc.checksum(path=test_file_path)
    assert check_sum == "etag"
    result = azc.isfile(path=test_file_path)
    assert result


@pytest.mark.parametrize("blob_client", [
    # blob client
    blob_client_credential,
    blob_client_connection_string,
])
def test_blob_info_error(mocker, blob_client):
    func_mock = mocker.MagicMock()
    func_mock.side_effect = IOError
    mocker.patch.object(BlobClient, "get_blob_properties", func_mock)

    result = azc.isfile(path=test_file_path)
    assert not result
    result = azc.isdir(path=test_file_path)
    assert not result


@pytest.mark.parametrize("blob_client", [
    # blob client
    blob_client_credential,
    blob_client_connection_string,
])
def test_blob_upload(mocker, blob_client):
    # ===================== #
    # test for AzBlobClient #
    # ===================== #

    # mock
    func_mock = mocker.MagicMock()
    func_mock.return_value = True

    mocker.patch.object(BlobClient, "upload_blob", func_mock)
    result = blob_client.put(path=test_file_path, data={})
    assert result


@pytest.mark.parametrize("blob_client", [
    # blob client
    blob_client_credential,
    blob_client_connection_string,
])
def test_blob_rm(mocker, blob_client):
    # ===================== #
    # test for AzBlobClient #
    # ===================== #

    # mock
    func_mock = mocker.MagicMock()
    func_mock.return_value = True

    mocker.patch.object(BlobClient, "delete_blob", func_mock)
    result = blob_client.rm(path=test_file_path)
    assert result


@pytest.mark.parametrize("blob_client", [
    # blob client
    blob_client_credential,
    blob_client_connection_string,
])
def test_blob_ls(mocker, blob_client):
    # ===================== #
    # test for AzBlobClient #
    # ===================== #

    # mock
    func_mock = mocker.MagicMock()
    func_mock.return_value = [
        BlobMock("test.csv"),
    ]

    mocker.patch.object(ContainerClient, "list_blobs", func_mock)
    file_list = blob_client.ls(path=test_file_ls_path, file_path=test_file_ls_path)
    assert file_list



from azure.cosmosdb.table.tableservice import TableService
import pytest

from azfs import TableStorage, TableStorageWrapper
from azfs.error import AzfsInputError


TABLE_NAME = "test_table"
PARTITION_KEY = "test_partition"
ROW_KEY = "test_row_key"
cons = {
    "account_name": "test",
    "account_key": "abcdefg1234",
    "database_name": TABLE_NAME,
}

EXAMPLE_MESSAGE = "test_message"

table_storage = TableStorage(**cons)
table_storage_wrapper = TableStorageWrapper(partition_key=PARTITION_KEY, **cons)


def test_table_storage_put(mocker):
    # data
    put_data = {"test_key": "test_value"}

    # mock
    func_mock = mocker.MagicMock()

    # patch
    mocker.patch.object(TableService, "insert_entity", func_mock)
    _ = table_storage.put(partition_key_value=PARTITION_KEY, data=put_data)


def test_table_storage_get(mocker):
    # mock
    func_mock = mocker.MagicMock()
    func_mock.return_value = []

    # patch
    mocker.patch.object(TableService, "query_entities", func_mock)
    _ = table_storage.get(partition_key_value=PARTITION_KEY, filter_key_values={})

    # filter value
    filter_value = f"PartitionKey eq '{PARTITION_KEY}'"
    func_mock.assert_called_with(table_name=TABLE_NAME, filter=filter_value)


def test_table_storage_update(mocker):
    #
    put_data = {"test_key": "test_value_2"}

    # mock
    func_mock = mocker.MagicMock()
    func_mock.return_value = []

    # patch
    mocker.patch.object(TableService, "update_entity", func_mock)
    _ = table_storage.update(partition_key_value=PARTITION_KEY, row_key=ROW_KEY, data=put_data)

    # filter value
    updated_data = {'PartitionKey': PARTITION_KEY, 'RowKey': ROW_KEY}
    updated_data.update(put_data)
    func_mock.assert_called_with(table_name=TABLE_NAME, entity=updated_data)


def test_table_storage_wrapper_put(mocker):
    # mock
    func_mock = mocker.MagicMock()

    # patch
    mocker.patch.object(TableService, "insert_entity", func_mock)
    _ = table_storage_wrapper.put(id_=ROW_KEY, message=EXAMPLE_MESSAGE)

    # validate
    put_data = {
        "PartitionKey": PARTITION_KEY,
        "RowKey": ROW_KEY,
        "message": EXAMPLE_MESSAGE
    }
    func_mock.assert_called_with(table_name=TABLE_NAME, entity=put_data)

    with pytest.raises(AzfsInputError):
        @table_storage_wrapper.overwrite_pack_data_to_put()
        def _pack_data(error_id: str):
            return {"id_": error_id}

    @table_storage_wrapper.overwrite_pack_data_to_put(allowed={"id_": [ROW_KEY]})
    def _pack_data(id_: str, message: str):
        return {"id_": id_, "message": message, "hash": f"{id_}-{message}"}

    _ = table_storage_wrapper.put(id_=ROW_KEY, message=EXAMPLE_MESSAGE)

    # validate
    put_data = {
        "PartitionKey": PARTITION_KEY,
        "RowKey": ROW_KEY,
        "message": EXAMPLE_MESSAGE,
        "hash": f"{ROW_KEY}-{EXAMPLE_MESSAGE}"
    }
    func_mock.assert_called_with(table_name=TABLE_NAME, entity=put_data)

    with pytest.raises(AzfsInputError):
        _ = table_storage_wrapper.put(id_="unexpected_row_key", message=EXAMPLE_MESSAGE)


def test_table_storage_wrapper_get(mocker):
    # mock
    func_mock = mocker.MagicMock()

    # patch
    mocker.patch.object(TableService, "query_entities", func_mock)
    _ = table_storage_wrapper.get(id_=ROW_KEY)

    # validate
    filter_value = f"PartitionKey eq '{PARTITION_KEY}' and RowKey eq '{ROW_KEY}'"
    func_mock.assert_called_with(table_name=TABLE_NAME, filter=filter_value)


def test_table_storage_wrapper_update(mocker):
    # mock
    func_mock = mocker.MagicMock()
    func_mock.return_value = []

    # patch
    mocker.patch.object(TableService, "update_entity", func_mock)
    _ = table_storage_wrapper.update(id_=ROW_KEY, message=EXAMPLE_MESSAGE)

    # validate
    updated_data = {
        "PartitionKey": PARTITION_KEY,
        "RowKey": ROW_KEY,
        "message": EXAMPLE_MESSAGE
    }
    func_mock.assert_called_with(table_name=TABLE_NAME, entity=updated_data)



import pytest
from azfs.clients.queue_client import AzQueueClient


credential = ""
queue_client = AzQueueClient(credential=credential)
test_queue_path = "https://test.queue.core.windows.net/test_queue/"


def test_not_implemented_error():
    with pytest.raises(NotImplementedError):
        queue_client.info(path=test_queue_path)

    with pytest.raises(NotImplementedError):
        queue_client.rm(path=test_queue_path)

    with pytest.raises(NotImplementedError):
        queue_client.get_container_client_from_path(path=test_queue_path)



import pytest
from azfs.utils import (
    BlobPathDecoder,
    ls_filter
)
from azfs.error import (
    AzfsInputError,
    AzfsInvalidPathError
)


class TestBlobPathDecoder:

    @pytest.mark.parametrize("path,storage_account_name,account_type,container_name,blob_file", [
        # blob storage
        ("https://test.blob.core.windows.net/test/test_file.csv", "test", "blob", "test", "test_file.csv"),
        ("https://test.blob.core.windows.net/test/dir/test_file.csv", "test", "blob", "test", "dir/test_file.csv"),
        ("https://test.blob.core.windows.net/test/", "test", "blob", "test", ""),
        ("https://test.blob.core.windows.net/test", "test", "blob", "test", ""),
        ("https://test.blob.core.windows.net/", "test", "blob", "", ""),
        ("blob/test/test/test_file.csv", "test", "blob", "test", "test_file.csv"),
        ("/blob/test/test/test_file.csv", "test", "blob", "test", "test_file.csv"),
        # datalake storage
        ("https://test.dfs.core.windows.net/test/test_file.csv", "test", "dfs", "test", "test_file.csv"),
        ("https://test.dfs.core.windows.net/test/dir/test_file.csv", "test", "dfs", "test", "dir/test_file.csv"),
        ("https://test.dfs.core.windows.net/test/", "test", "dfs", "test", ""),
        ("https://test.dfs.core.windows.net/test", "test", "dfs", "test", ""),
        ("https://test.dfs.core.windows.net/", "test", "dfs", "", ""),
        ("dfs/test/test/test_file.csv", "test", "dfs", "test", "test_file.csv"),
        ("/dfs/test/test/test_file.csv", "test", "dfs", "test", "test_file.csv"),
    ])
    def test_path_decoder_pass(self, path, storage_account_name, account_type, container_name, blob_file):
        bpd = BlobPathDecoder()
        storage_account_name_v, account_type_v, container_name_v, blob_file_v = bpd.decode(path).get()
        assert storage_account_name_v == storage_account_name
        assert account_type_v == account_type
        assert container_name_v == container_name
        assert blob_file_v == blob_file

        storage_account_name_v, account_type_v, container_name_v, blob_file_v = BlobPathDecoder(path).get()
        assert storage_account_name_v == storage_account_name
        assert account_type_v == account_type
        assert container_name_v == container_name
        assert blob_file_v == blob_file

    def test_path_decoder_error(self):
        path = "https://aaa/bbb/ccc"

        bpd = BlobPathDecoder()
        with pytest.raises(AzfsInvalidPathError):
            bpd.decode(path)

        with pytest.raises(AzfsInvalidPathError):
            BlobPathDecoder(path)

    @pytest.mark.parametrize("path_pattern,path,storage_account_name,account_type,container_name,blob_file", [
        # blob storage
        ("%A=%T=%C/%B", "test=blob=test/test_file.csv", "test", "blob", "test", "test_file.csv"),
        ("%A=%T=%C/%B", "test=blob=test/dir/test_file.csv", "test", "blob", "test", "dir/test_file.csv"),
        # datalake storage
        ("%A=%T=%C/%B", "test=dfs=test/test_file.csv", "test", "dfs", "test", "test_file.csv"),
        ("%A=%T=%C/%B", "test=dfs=test/dir/test_file.csv", "test", "dfs", "test", "dir/test_file.csv"),
    ])
    def test_add_pattern(self, path_pattern, path, storage_account_name, account_type, container_name, blob_file):
        # add new pattern
        BlobPathDecoder.add_pattern(pattern=path_pattern)
        storage_account_name_v, account_type_v, container_name_v, blob_file_v = BlobPathDecoder(path).get()
        assert storage_account_name_v == storage_account_name
        assert account_type_v == account_type
        assert container_name_v == container_name
        assert blob_file_v == blob_file

    def test_add_pattern_error(self):
        # shortage of %
        path_pattern = "%A=%T=%C"
        with pytest.raises(AzfsInputError):
            BlobPathDecoder.add_pattern(pattern=path_pattern)

        # unknown %
        path_pattern = "%A=%T=%C%Z"
        with pytest.raises(AzfsInputError):
            BlobPathDecoder.add_pattern(pattern=path_pattern)


class TestLsFilter:
    def test_ls_filter(self):
        file_path_list = [
            "file1.csv",
            "file2.csv",
            "file3.csv",
            "dir/file4.csv",
            "dir/file5.csv",
            "dir/some/file6.csv"
        ]
        # sorted by alphabetic order
        result_list = [
            "dir/",
            "file1.csv",
            "file2.csv",
            "file3.csv"
        ]
        result_list_v = ls_filter(file_path_list, "")
        assert result_list_v == result_list

        result_list = [
            "file4.csv",
            "file5.csv",
            "some/"
        ]
        result_list_v = ls_filter(file_path_list, "dir")
        assert result_list_v == result_list

        result_list = [
            "file6.csv"
        ]
        result_list_v = ls_filter(file_path_list, "dir/some")
        assert result_list_v == result_list



import bz2
from functools import partial
import gzip
import io
import json
import lzma
import multiprocessing as mp
import pickle
import re
from typing import Union, Optional, List
import warnings

import pandas as pd
from azure.identity import DefaultAzureCredential
from azure.core.exceptions import ResourceNotFoundError
from azfs.clients import AzfsClient, TextReader
from azfs.error import AzfsInputError
from azfs.utils import (
    BlobPathDecoder,
    ls_filter
)

__all__ = ["AzFileClient"]


def _wrap_quick_load(inputs: dict):
    """
    read wrapper function for multiprocessing.

    Args:
        inputs:

    Returns:

    """
    return _quick_load(**inputs)


def _quick_load(
        path: str,
        file_format: Optional[str] = None,
        credential: Optional[str] = None,
        apply_method: Optional[callable] = None) -> pd.DataFrame:
    """
    read function for multiprocessing.

    Args:
        path: file-path to read
        file_format: format of the file
        credential:
        apply_method:

    Returns:
        pd.DataFrame
    """
    if credential is None:
        azc = AzFileClient()
    else:
        azc = AzFileClient(credential=credential)

    # set file_format if None
    if file_format is None:
        if path.endswith(".csv"):
            file_format = "csv"
        elif path.endswith(".parquet"):
            file_format = "parquet"
        elif path.endswith(".pkl"):
            file_format = "pickle"
        else:
            raise AzfsInputError("file_format is incorrect")

    # read file as pandas DataFrame
    if file_format == "csv":
        df = azc.read_csv(path=path)
    elif file_format == "parquet":
        df = azc.read_parquet(path=path)
    elif file_format == "pickle":
        df = azc.read_pickle(path=path)
    else:
        raise AzfsInputError("file_format is incorrect")

    # apply additional function
    if apply_method is None:
        return df
    else:
        return apply_method(df)


class DataFrameReader:
    def __init__(
            self,
            _azc,
            credential: Union[str, DefaultAzureCredential],
            path: Union[str, List[str]] = None,
            use_mp=False,
            cpu_count: Optional[int] = None,
            file_format: Optional[str] = None):
        self._azc: AzFileClient = _azc
        # DefaultCredential cannot be pickle (when use multiprocessing), so make it None
        self._credential = credential if type(credential) is str else None
        self.path: Optional[List[str]] = self._decode_path(path=path)
        self.file_format = file_format
        self.use_mp = use_mp
        self.cpu_count = mp.cpu_count() if cpu_count is None else cpu_count
        self._apply_method = None

    def _decode_path(self, path: Optional[Union[str, List[str]]]) -> Optional[List[str]]:
        """
        decode path to be read by azc

        Args:
            path: azure blob path

        Returns:

        """
        if path is None:
            return None
        elif type(path) is str:
            if "*" in path:
                decoded_path = self._azc.glob(pattern_path=path)
            else:
                decoded_path = [path]
        elif type(path) is list:
            decoded_path = path
        else:
            raise AzfsInputError("path must be `str` or `list`")
        return decoded_path

    def csv(self, path: Union[str, List[str]] = None, **kwargs) -> pd.DataFrame:
        """
        read csv files in Azure Blob, like PySpark-method.

        Args:
            path: azure blob path
            **kwargs: as same as pandas.read_csv

        Returns:
            pd.DataFrame

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> blob_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            >>> df = azc.read().csv(blob_path)
            # result is as same as azc.read_csv(blob_path)
            >>> blob_path_list = [
            ...     "https://testazfs.blob.core.windows.net/test_container/test1.csv",
            ...     "https://testazfs.blob.core.windows.net/test_container/test2.csv"
            ... ]
            >>> df = azc.read().csv(blob_path_list)
            # result is as same as pd.concat([each data-frame])
            # in addition, you can use `*`
            >>> blob_path_pattern = "https://testazfs.blob.core.windows.net/test_container/test*.csv"
            >>> df = azc.read().csv(blob_path_pattern)
            # you can use multiprocessing with `use_mp` argument
            >>> df = azc.read(use_mp=True).csv(blob_path_pattern)
            # if you want to filter or apply some method, you can use your defined function as below
            >>> def filter_function(_df: pd.DataFrame, _id: str) -> pd.DataFrame:
            ...     return _df[_df['id'] == _id]
            >>> df = azc.read(use_mp=True).apply(function=filter_function, _id="aaa").csv(blob_path_pattern)


        """
        self.file_format = "csv"
        if path is not None:
            self.path = self._decode_path(path=path)
        return self._load(**kwargs)

    def parquet(self, path: Union[str, List[str]] = None) -> pd.DataFrame:
        """
        read parquet files in Azure Blob, like PySpark-method.

        Args:
            path: azure blob path

        Returns:
            pd.DataFrame

        """
        self.file_format = "parquet"
        if path is not None:
            self.path = self._decode_path(path=path)
        return self._load()

    def pickle(self, path: Union[str, List[str]] = None, compression: str = "gzip") -> pd.DataFrame:
        """
        read pickle files in Azure Blob, like PySpark-method.

        Args:
            path: azure blob path
            compression: acceptable keywords are: gzip, bz2, xz. gzip is default value.

        Returns:
            pd.DataFrame

        """
        self.file_format = "pickle"
        if path is not None:
            self.path = self._decode_path(path=path)
        return self._load(compression=compression)

    def _load_function(self) -> callable:
        """
        get read_* function according to the file_format

        Returns:

        """
        if self.file_format == "csv":
            load_function = self._azc.read_csv
        elif self.file_format == "parquet":
            load_function = self._azc.read_parquet
        elif self.file_format == "pickle":
            load_function = self._azc.read_pickle
        else:
            raise AzfsInputError("file_format is incorrect")
        return load_function

    def apply(self, *, function: callable, **kwargs):
        """
        to apply pandas DataFrame

        Args:
            function: first argument must pass pd.DataFrame
            **kwargs: argument to pass the function

        Returns:
            self
        """
        if kwargs:
            self._apply_method = partial(function, **kwargs)
        else:
            self._apply_method = function
        return self

    def _load(self, **kwargs) -> Optional[pd.DataFrame]:
        if self.path is None:
            raise AzfsInputError("input azure blob path")

        if self.use_mp:
            params_list = []
            for f in self.path:
                _input = {
                    "path": f,
                    "file_format": self.file_format,
                    "credential": self._credential,
                    "apply_method": self._apply_method
                }
                _input.update(kwargs)
                params_list.append(_input)
            with mp.Pool(self.cpu_count) as pool:
                df_list = pool.map(_wrap_quick_load, params_list)
            pool.join()
        else:
            load_function = self._load_function()
            if self._apply_method is None:
                df_list = [load_function(f, **kwargs) for f in self.path]
            else:
                df_list = [self._apply_method(load_function(f, **kwargs)) for f in self.path]
        if len(df_list) == 0:
            return None
        return pd.concat(df_list)


class AzFileClient:
    """

    AzFileClient is

    * list files in blob (also with wildcard ``*``),
    * check if file exists,
    * read csv as pd.DataFrame, and json as dict from blob,
    * write pd.DataFrame as csv, and dict as json to blob,

    Examples:
        >>> import azfs
        >>> from azure.identity import DefaultAzureCredential
        credential is not required if your environment is on AAD
        >>> azc = azfs.AzFileClient()
        credential is required if your environment is not on AAD
        >>> credential = "[your storage account credential]"
        >>> azc = azfs.AzFileClient(credential=credential)
        # or
        >>> credential = DefaultAzureCredential()
        >>> azc = azfs.AzFileClient(credential=credential)
        connection_string will be also acceptted
        >>> connection_string = "[your connection_string]"
        >>> azc = azfs.AzFileClient(connection_string=connection_string)
    """

    class AzContextManager:
        """
        AzContextManger provides easy way to set new function as attribute to another package like pandas.
        """
        def __init__(self):
            self.register_list = []

        def register(self, _as: str, _to: object):
            """
            register decorated function to self.register_list.


            Args:
                _as: new method name
                _to: assign to class or object

            Returns:
                decorated function

            """
            def _register(function):
                """
                append ``wrapper`` function

                Args:
                    function:

                Returns:

                """
                def wrapper(class_instance):
                    """
                    accept instance in kwargs as name of ``az_file_client_instance``

                    Args:
                        class_instance: always instance of AzFileClient

                    Returns:

                    """

                    def new_function(*args, **kwargs):
                        """
                        actual wrapped function

                        Args:
                            *args:
                            **kwargs:

                        Returns:

                        """
                        target_function = getattr(class_instance, function.__name__)

                        df = args[0] if isinstance(args[0], pd.DataFrame) else None
                        if df is not None:
                            kwargs['df'] = args[0]
                            return target_function(*args[1:], **kwargs)
                        return target_function(*args, **kwargs)

                    return new_function

                function_info = {
                    "assign_as": _as,
                    "assign_to": _to,
                    "function": wrapper
                }
                self.register_list.append(function_info)

                return function

            return _register

        def attach(self, client: object):
            """
            set new function as attribute based on self.register_list

            Args:
                client: set AzFileClient always

            Returns:
                None

            """
            for f in self.register_list:
                setattr(f['assign_to'], f['assign_as'], f['function'](class_instance=client))

        def detach(self):
            """
            set None based on self.register_list

            Returns:
                None

            """
            for f in self.register_list:
                setattr(f['assign_to'], f['assign_as'], None)

    # instance for context manager
    _az_context_manager = AzContextManager()

    def __init__(
            self,
            credential: Optional[Union[str, DefaultAzureCredential]] = None,
            connection_string: Optional[str] = None):
        """
        if every argument is None, set credential as DefaultAzureCredential().

        Args:
            credential: if string, Blob Storage -> Access Keys -> Key
            connection_string: connection_string
        """

        if credential is None and connection_string is None:
            credential = DefaultAzureCredential()
        self._client = AzfsClient(credential=credential, connection_string=connection_string)
        self._credential = credential

    def __enter__(self):
        """
        add some functions to pandas module based on AzContextManger()

        Returns:
            instance of AzFileClient

        """
        self._az_context_manager.attach(client=self)
        return self

    def __exit__(self, exec_type, exec_value, traceback):
        """
        remove some functions from pandas module based on AzContextManager()

        Args:
            exec_type:
            exec_value:
            traceback:

        Returns:
            None
        """
        self._az_context_manager.detach()

    def exists(self, path: str) -> bool:
        """
        check if specified file exists or not.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``

        Returns:
            ``True`` if files exists, otherwise ``False``

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            >>> azc.exists(path=csv_path)
            True
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/not_exist_test1.csv"
            >>> azc.exists(path=csv_path)
            False

        """
        try:
            _ = self.info(path=path)
        except ResourceNotFoundError:
            return False
        else:
            return True

    def ls(self, path: str, attach_prefix: bool = False) -> list:
        """
        list blob file from blob or dfs.

        Args:
            path: Azure Blob path URL format, ex: https://testazfs.blob.core.windows.net/test_container
            attach_prefix: return full_path if True, return only name

        Returns:
            list of azure blob files

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container"
            >>> azc.ls(csv_path)
            [
                "test1.csv",
                "test2.csv",
                "test3.csv",
                "directory_1",
                "directory_2"
            ]
            >>> azc.ls(path=path, attach_prefix=True)
            [
                "https://testazfs.blob.core.windows.net/test_container/test1.csv",
                "https://testazfs.blob.core.windows.net/test_container/test2.csv",
                "https://testazfs.blob.core.windows.net/test_container/test3.csv",
                "https://testazfs.blob.core.windows.net/test_container/directory_1",
                "https://testazfs.blob.core.windows.net/test_container/directory_2"
            ]

        """
        _, account_kind, _, file_path = BlobPathDecoder(path).get_with_url()
        file_list = self._client.get_client(account_kind=account_kind).ls(path=path, file_path=file_path)
        if account_kind in ["dfs", "blob"]:
            file_name_list = ls_filter(file_path_list=file_list, file_path=file_path)
            if attach_prefix:
                path = path if path.endswith("/") else f"{path}/"
                file_full_path_list = [f"{path}{f}" for f in file_name_list]
                return file_full_path_list
            else:
                return file_name_list
        elif account_kind in ["queue"]:
            return file_list

    def cp(self, src_path: str, dst_path: str, overwrite=False) -> bool:
        """
        copy the data from `src_path` to `dst_path`

        Args:
            src_path:
                Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``
            dst_path:
                Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test2.csv``
            overwrite:

        Returns:

        """
        if src_path == dst_path:
            raise AzfsInputError("src_path and dst_path must be different")
        if (not overwrite) and self.exists(dst_path):
            raise AzfsInputError(f"{dst_path} is already exists. Please set `overwrite=True`.")
        data = self._get(path=src_path)
        if type(data) is io.BytesIO:
            self._put(path=dst_path, data=data.read())
        elif type(data) is bytes:
            self._put(path=dst_path, data=data)
        return True

    def rm(self, path: str) -> bool:
        """
        delete the file in blob

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``

        Returns:
            True if target file is correctly removed.

        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        return self._client.get_client(account_kind=account_kind).rm(path=path)

    def info(self, path: str) -> dict:
        """
        get file properties, such as
        ``name``,  ``creation_time``, ``last_modified_time``, ``size``, ``content_hash(md5)``.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``

        Returns:
            dict info of some file

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            >>> azc.info(path=csv_path)
            {
                "name": "test1.csv",
                "size": "128KB",
                "creation_time": "",
                "last_modified": "",
                "etag": "etag...",
                "content_type": "",
                "type": "file"
            }

        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        # get info from blob or data-lake storage
        data = self._client.get_client(account_kind=account_kind).info(path=path)

        # extract below to determine file or directory
        content_settings = data.get("content_settings", {})
        metadata = data.get("metadata", {})

        data_type = ""
        if "hdi_isfolder" in metadata:
            # only data-lake storage has `hdi_isfolder`
            data_type = "directory"
        elif content_settings.get("content_type") is not None:
            # blob and data-lake storage have `content_settings`,
            # and its value of the `content_type` must not be None
            data_type = "file"
        return {
            "name": data.get("name", ""),
            "size": data.get("size", ""),
            "creation_time": data.get("creation_time", ""),
            "last_modified": data.get("last_modified", ""),
            "etag": data.get("etag", ""),
            "content_type": content_settings.get("content_type", ""),
            "type": data_type
        }

    def checksum(self, path: str) -> str:
        """
        Blob and DataLake storage have etag.

        Args:
            path:

        Returns:
            etag

        Raises:
            KeyError: if info has no etag

        """
        return self.info(path=path)["etag"]

    def size(self, path) -> Optional[Union[int, str]]:
        """
        Size in bytes of file

        Args:
            path:

        Returns:

        """
        return self.info(path).get("size")

    def isdir(self, path) -> bool:
        """
        Is this entry directory-like?

        Args:
            path:

        Returns:

        """
        try:
            return self.info(path)["type"] == "directory"
        except IOError:
            return False

    def isfile(self, path) -> bool:
        """
        Is this entry file-like?

        Args:
            path:

        Returns:

        """
        try:
            return self.info(path)["type"] == "file"
        except IOError:
            return False

    def glob(self, pattern_path: str) -> List[str]:
        """
        Currently only support ``* (wildcard)`` .
        By default, ``glob()`` lists specified files with formatted-URL.

        Args:
            pattern_path: ex: ``https://<storage_account_name>.blob.core.windows.net/<container>/*/*.csv``

        Returns:
            lists specified files filtered by wildcard

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> path = "https://testazfs.blob.core.windows.net/test_container/some_folder"
            ls() lists all files in some folder like
            >>> azc.ls(path)
            [
                "test1.csv",
                "test2.csv",
                "test3.csv",
                "test1.json",
                "test2.json",
                "directory_1",
                "directory_2"
            ]
            glob() lists specified files according to the wildcard, and lists with formatted-URL by default
            >>> csv_pattern_path = "https://testazfs.blob.core.windows.net/test_container/some_folder/*.csv"
            >>> azc.glob(path=csv_pattern_path)
            [
                "https://testazfs.blob.core.windows.net/test_container/some_folder/test1.csv",
                "https://testazfs.blob.core.windows.net/test_container/some_folder/test2.csv",
                "https://testazfs.blob.core.windows.net/test_container/some_folder/test3.csv"
            ]
            glob() can use any path
            >>> csv_pattern_path = "https://testazfs.blob.core.windows.net/test_container/some_folder/test1.*"
            >>> azc.glob(path=csv_pattern_path)
            [
                "https://testazfs.blob.core.windows.net/test_container/some_folder/test1.csv",
                "https://testazfs.blob.core.windows.net/test_container/some_folder/test1.json"
            ]
            also deeper folders
            >>> csv_pattern_path = "https://testazfs.blob.core.windows.net/test_container/some_folder/*/*.csv"
            >>> azc.glob(path=csv_pattern_path)
            [
                "https://testazfs.blob.core.windows.net/test_container/some_folder/directory_1/deeper_test1.csv",
                "https://testazfs.blob.core.windows.net/test_container/some_folder/directory_2/deeper_test2.csv"
            ]

        Raises:
            AzfsInputError: when ``*`` is used in root_flder under a container.
        """
        if "*" not in pattern_path:
            raise AzfsInputError("no any `*` in the `pattern_path`")
        url, account_kind, container_name, file_path = BlobPathDecoder(pattern_path).get_with_url()

        acceptable_folder_pattern = r"(?P<root_folder>[^\*]+)/(?P<folders>.*)"
        result = re.match(acceptable_folder_pattern, file_path)
        if result:
            result_dict = result.groupdict()
            root_folder = result_dict['root_folder']
        else:
            raise AzfsInputError(
                f"Cannot use `*` in root_folder under a container. Accepted format is {acceptable_folder_pattern}"
            )
        # get container root path
        base_path = f"{url}/{container_name}/"
        if account_kind in ["dfs", "blob"]:
            file_list = self._client.get_client(account_kind=account_kind).ls(path=base_path, file_path=root_folder)

            # to escape special chars for regular-expression
            def _escape(input_str: str) -> str:
                special_chars = ["(", ")", "[", "]"]
                for c in special_chars:
                    input_str = input_str.replace(c, f"\\{c}")
                return input_str

            escaped_pattern_path = _escape(pattern_path)
            # fix pattern_path, in order to avoid matching `/`
            replace_pattern_path = escaped_pattern_path.replace('*', '([^/])*?')
            pattern = re.compile(f"{replace_pattern_path}$")
            file_full_path_list = [f"{base_path}{f}" for f in file_list]
            # filter with pattern.match
            matched_full_path_list = [f for f in file_full_path_list if pattern.match(f)]
            return matched_full_path_list
        elif account_kind in ["queue"]:
            raise NotImplementedError

    def read(
            self,
            *,
            path: Union[str, List[str]] = None,
            use_mp: bool = False,
            cpu_count: Optional[int] = None,
            file_format: str = "csv") -> DataFrameReader:
        """
        read csv, parquet, picke files in Azure Blob, like PySpark-method.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``
            use_mp: Default, False
            cpu_count: Default, as same as mp.cpu_count()
            file_format: determined by which function you call

        Returns:
            pd.DataFrame

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> blob_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            >>> df = azc.read().csv(blob_path)
            # result is as same as azc.read_csv(blob_path)
            >>> blob_path_list = [
            ...     "https://testazfs.blob.core.windows.net/test_container/test1.csv",
            ...     "https://testazfs.blob.core.windows.net/test_container/test2.csv"
            ... ]
            >>> df = azc.read().csv(blob_path_list)
            # result is as same as pd.concat([each data-frame])
            # in addition, you can use `*`
            >>> blob_path_pattern = "https://testazfs.blob.core.windows.net/test_container/test*.csv"
            >>> df = azc.read().csv(blob_path_pattern)
            # you can use multiprocessing with `use_mp` argument
            >>> df = azc.read(use_mp=True).csv(blob_path_pattern)
            # if you want to filter or apply some method, you can use your defined function as below
            >>> def filter_function(_df: pd.DataFrame, _id: str) -> pd.DataFrame:
            ...     return _df[_df['id'] == _id]
            >>> df = azc.read(use_mp=True).apply(function=filter_function, _id="aaa").csv(blob_path_pattern)


        """
        return DataFrameReader(
            _azc=self,
            credential=self._credential,
            path=path,
            use_mp=use_mp,
            cpu_count=cpu_count,
            file_format=file_format)

    def _get(self, path: str, offset: int = None, length: int = None, **kwargs) -> Union[bytes, str, io.BytesIO, dict]:
        """
        get data from Azure Blob Storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``
            offset:
            length:
            **kwargs:

        Returns:
            some data

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            you can read csv file in azure blob storage
            >>> data = azc.get(path=csv_path)
            `download()` is same method as `get()`
            >>> data = azc.download(path=csv_path)

        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()

        file_bytes = self._client.get_client(
            account_kind=account_kind).get(path=path, offset=offset, length=length, **kwargs)
        # gzip圧縮ファイルは一旦ここで展開
        if path.endswith(".gz"):
            file_bytes = gzip.decompress(file_bytes)

        if type(file_bytes) is bytes:
            file_to_read = io.BytesIO(file_bytes)
        else:
            file_to_read = file_bytes

        return file_to_read

    def read_line_iter(self, path: str) -> iter:
        """
        To read text file in each line with iterator.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``

        Returns:
            get data of the path as iterator

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            >>> for l in azc.read_line_iter(path=csv_path)
            ...     print(l.decode("utf-8"))

        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        return TextReader(client=self._client.get_client(account_kind=account_kind), path=path)

    def read_csv_chunk(self, path: str, chunk_size: int) -> pd.DataFrame:
        """
        !WARNING! the method may differ from current version in the future update.
        Currently, only support for csv.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``
            chunk_size: pandas-DataFrame index length to read.

        Returns:
            first time: len(df.index) is `chunk_size - 1`
            second time or later: len(df.index) is `chunk_size`

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            >>> read_chunk_size = 100
            >>> for _df in azc.read_csv_chunk(path=csv_path, chunk_size=read_chunk_size):
            ...   print(_df)
        """
        warning_message = """
            The method is under developing. 
            The name or the arguments may differ from current version in the future update.
        """
        warnings.warn(warning_message, FutureWarning)
        initial_line = ""
        byte_list = []

        for idx, l in enumerate(self.read_line_iter(path=path)):
            div_idx = idx % chunk_size
            if idx == 0:
                initial_line = l
                byte_list.append(initial_line)
            else:
                byte_list.append(l)
            if div_idx + 1 == chunk_size:
                file_to_read = (b"\n".join(byte_list))
                file_to_io_read = io.BytesIO(file_to_read)
                df = pd.read_csv(file_to_io_read)
                yield df

                byte_list = [initial_line]
        # make remainder DataFrame after the for-loop
        file_to_read = (b"\n".join(byte_list))
        file_to_io_read = io.BytesIO(file_to_read)
        df = pd.read_csv(file_to_io_read)
        yield df

    @_az_context_manager.register(_as="read_csv_az", _to=pd)
    def read_csv(self, path: str, **kwargs) -> pd.DataFrame:
        """
        get csv data as pd.DataFrame from Azure Blob Storage.
        support ``csv`` and also ``csv.gz``.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``
            **kwargs: keywords to put df.read_csv(), such as ``header``, ``encoding``.

        Returns:
            pd.DataFrame

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            you can read and write csv file in azure blob storage
            >>> df = azc.read_csv(path=csv_path)
            Using `with` statement, you can use `pandas`-like methods
            >>> with azc:
            >>>     df = pd.read_csv_az(path)

        """
        file_to_read = self._get(path)
        return pd.read_csv(file_to_read, **kwargs)

    @_az_context_manager.register(_as="read_table_az", _to=pd)
    def read_table(self, path: str, **kwargs) -> pd.DataFrame:
        """
        get tsv data as pd.DataFrame from Azure Blob Storage.
        support ``tsv``.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.tsv``
            **kwargs: keywords to put df.read_csv(), such as ``header``, ``encoding``.

        Returns:
            pd.DataFrame

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> tsv_path = "https://testazfs.blob.core.windows.net/test_container/test1.tsv"
            you can read and write csv file in azure blob storage
            >>> df = azc.read_table(path=tsv_path)
            Using `with` statement, you can use `pandas`-like methods
            >>> with azc:
            >>>     df = pd.read_table_az(tsv_path)

        """
        file_to_read = self._get(path)
        return pd.read_table(file_to_read, **kwargs)

    @_az_context_manager.register(_as="read_pickle_az", _to=pd)
    def read_pickle(self, path: str, compression="gzip") -> pd.DataFrame:
        """
        get pickled-pandas data as pd.DataFrame from Azure Blob Storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.pkl``
            compression: acceptable keywords are: gzip, bz2, xz. gzip is default value.

        Returns:
            pd.DataFrame

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> pkl_path = "https://testazfs.blob.core.windows.net/test_container/test1.pkl"
            you can read and write csv file in azure blob storage
            >>> df = azc.read_pickle(path=pkl_path)
            Using `with` statement, you can use `pandas`-like methods
            >>> with azc:
            >>>     df = pd.read_pickle_az(pkl_path)
            you can use difference compression
            >>> with azc:
            >>>     df = pd.read_pickle_az(pkl_path, compression="bz2")

        """
        file_to_read = self._get(path).read()
        if compression == "gzip":
            file_to_read = gzip.decompress(file_to_read)
        elif compression == "bz2":
            file_to_read = bz2.decompress(file_to_read)
        elif compression == "xz":
            file_to_read = lzma.decompress(file_to_read)
        return pd.DataFrame(pickle.loads(file_to_read))

    @_az_context_manager.register(_as="read_parquet_az", _to=pd)
    def read_parquet(self, path: str) -> pd.DataFrame:
        """

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test.parquet``

        Returns:
            pd.DataFrame

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> parquet_path = "https://testazfs.blob.core.windows.net/test_container/test1.parquet"
            you can read and write csv file in azure blob storage
            >>> df = azc.read_parquet(path=parquet_path)
            Using `with` statement, you can use `pandas`-like methods
            >>> with azc:
            >>>     df = pd.read_parquet_az(parquet_path)


        """
        import pyarrow.parquet as pq
        data = self._get(path=path)
        return pq.read_table(data).to_pandas()

    def _put(self, path: str, data) -> bool:
        """
        upload data to blob or data_lake storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``
            data: some data to upload.

        Returns:
            True if correctly uploaded

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            you can write file in azure blob storage
            >>> _data = azc.put(path=csv_path)
            `download()` is same method as `get()`
            >>> _data = azc.upload(path=csv_path)

        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        return self._client.get_client(account_kind=account_kind).put(path=path, data=data)

    @_az_context_manager.register(_as="to_csv_az", _to=pd.DataFrame)
    def write_csv(self, path: str, df: pd.DataFrame, **kwargs) -> bool:
        """
        output pandas dataframe to csv file in Datalake storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.csv``.
            df: pd.DataFrame to upload.
            **kwargs: keywords to put df.to_csv(), such as ``encoding``, ``index``.

        Returns:
            True if correctly uploaded

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> csv_path = "https://testazfs.blob.core.windows.net/test_container/test1.csv"
            you can read and write csv file in azure blob storage
            >>> azc.write_csv(path=csv_path, df=df)
            Using `with` statement, you can use `pandas`-like methods
            >>> with azc:
            >>>     df.to_csv_az(csv_path)
        """
        csv_str = df.to_csv(**kwargs).encode("utf-8")
        return self._put(path=path, data=csv_str)

    @_az_context_manager.register(_as="to_table_az", _to=pd.DataFrame)
    def write_table(self, path: str, df: pd.DataFrame, **kwargs) -> bool:
        """
        output pandas dataframe to tsv file in Datalake storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.tsv``.
            df: pd.DataFrame to upload.
            **kwargs: keywords to put df.to_csv(), such as ``encoding``, ``index``.

        Returns:
            True if correctly uploaded

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> tsv_path = "https://testazfs.blob.core.windows.net/test_container/test1.tsv"
            you can read and write csv file in azure blob storage
            >>> azc.write_table(path=tsv_path, df=df)
            Using `with` statement, you can use `pandas`-like methods
            >>> with azc:
            >>>     df.to_table_az(tsv_path)
        """
        table_str = df.to_csv(sep="\t", **kwargs).encode("utf-8")
        return self._put(path=path, data=table_str)

    @_az_context_manager.register(_as="to_pickle_az", _to=pd.DataFrame)
    def write_pickle(self, path: str, df: pd.DataFrame, compression="gzip") -> bool:
        """
        output pandas dataframe to tsv file in Datalake storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.pkl``
            df: pd.DataFrame to upload.
            compression: acceptable keywords are: gzip, bz2, xz. gzip is default value.

        Returns:
            pd.DataFrame

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> pkl_path = "https://testazfs.blob.core.windows.net/test_container/test1.pkl"
            you can read and write csv file in azure blob storage
            >>> azc.write_pickle(path=pkl_path, df=df)
            Using `with` statement, you can use `pandas`-like methods
            >>> with azc:
            >>>     df.to_pickle_az(pkl_path)
            you can use difference compression
            >>> with azc:
            >>>     df.to_pickle_az(pkl_path, compression="bz2")

        """
        serialized_data = pickle.dumps(df)
        if compression == "gzip":
            serialized_data = gzip.compress(serialized_data)
        elif compression == "bz2":
            serialized_data = bz2.compress(serialized_data)
        elif compression == "xz":
            serialized_data = lzma.compress(serialized_data)
        return self._put(path=path, data=serialized_data)

    @_az_context_manager.register(_as="to_parquet_az", _to=pd.DataFrame)
    def write_parquet(self, path: str, table) -> bool:
        """
        When implementation of AzFileSystem is done, the function will be implemented.


        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test.parquet``
            table: parquet table

        Returns:
            True: if successfully uploaded

        Examples:
            >>> from azfs import AzFileSystem
            >>> import pyarrow.parquet as pq
            >>> fs = AzFileSystem()
            >>> with fs.open("azure_path", "wb") as f:
            ...     pq.write_table(table, f)

        """
        raise NotImplementedError

    def read_json(self, path: str, **kwargs) -> dict:
        """
        read json file in Datalake storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.json``
            **kwargs: keywords to put json.loads(), such as ``parse_float``.

        Returns:
            dict

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> json_path = "https://testazfs.blob.core.windows.net/test_container/test1.json"
            you can read and write csv file in azure blob storage
            >>> azc.read_json(path=json_path)

        """
        file_bytes = self._get(path)
        if type(file_bytes) is io.BytesIO:
            file_bytes = file_bytes.read()
        return json.loads(file_bytes, **kwargs)

    def write_json(self, path: str, data: dict, **kwargs) -> bool:
        """
        output dict to json file in Datalake storage.

        Args:
            path: Azure Blob path URL format, ex: ``https://testazfs.blob.core.windows.net/test_container/test1.json``
            data: dict to upload
            **kwargs: keywords to put json.loads(), such as ``indent``.

        Returns:
            True if correctly uploaded

        Examples:
            >>> import azfs
            >>> azc = azfs.AzFileClient()
            >>> json_path = "https://testazfs.blob.core.windows.net/test_container/test1.json"
            you can read and write csv file in azure blob storage
            >>> azc.write_json(path=json_path, data={"": ""})

        """
        # encode with UTF-8 to fully upload data including not ascii character
        return self._put(path=path, data=json.dumps(data, **kwargs).encode("utf-8"))

    # ===================
    # alias for functions
    # ===================

    get = _get
    get.__doc__ = _get.__doc__
    download = _get
    download.__doc__ = _get.__doc__
    put = _put
    put.__doc__ = _put.__doc__
    upload = _put
    upload.__doc__ = _put.__doc__



from abc import abstractmethod
from typing import Union, Optional
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobClient, ContainerClient
from azure.storage.filedatalake import DataLakeFileClient, FileSystemClient
from azure.storage.queue import QueueClient
from azfs.utils import BlobPathDecoder


# define the types
FileClientType = Union[
    BlobClient,
    DataLakeFileClient,
    QueueClient
]
FileSystemClientType = Union[
    ContainerClient,
    FileSystemClient
]


class ClientInterface:
    """
    The class provides Azure Blob, DataLake and Queue Client interface.
    Abstract methods below are implemented in each inherited classes.

    * _get_file_client
    * _get_service_client
    * _get_container_client
    * _ls
    * _get
    * _put
    """

    def __init__(
            self,
            credential: Optional[Union[str, DefaultAzureCredential]],
            connection_string: Optional[str] = None):
        self.credential = credential
        self.connection_string = connection_string

    @abstractmethod
    def _get_service_client_from_credential(
            self,
            account_url: str,
            credential: Union[DefaultAzureCredential, str]):
        """
        get service_client for Blob, DataLake or Queue Service Client with credential.

        Args:
            account_url: account_url ends with ``/``
            credential: DefaultCredential or string

        Returns:

        """
        raise NotImplementedError

    @abstractmethod
    def _get_service_client_from_connection_string(
            self,
            connection_string: str):
        raise NotImplementedError

    def _get_service_client_from_url(self, account_url):
        if self.credential is not None:
            return self._get_service_client_from_credential(account_url=account_url, credential=self.credential)
        elif self.connection_string is not None:
            return self._get_service_client_from_connection_string(connection_string=self.connection_string)

    def get_service_client_from_url(self, account_url):
        return self._get_service_client_from_url(account_url=account_url)

    def get_file_client_from_path(self, path: str) -> FileClientType:
        """
        get file_client from given path

        Args:
            path: Azure path that ``BlobPathDecode()`` can decode

        Returns:
            Union[BlobClient, DataLakeFileClient, QueueClient]
        """
        account_url, account_kind, file_system, file_path = BlobPathDecoder(path).get_with_url()
        return self._get_file_client(
            account_url=account_url,
            file_system=file_system,
            file_path=file_path)

    @abstractmethod
    def _get_file_client(
            self,
            account_url: str,
            file_system: str,
            file_path: str):
        """
        abstract method to be implemented
        get file_client from given path
        :param account_url:
        :param file_system:
        :param file_path:
        :return:
        """
        raise NotImplementedError

    def get_container_client_from_path(self, path: str) -> FileSystemClientType:
        """
        get container_client from given path

        Args:
            path: Azure path that ``BlobPathDecode()`` can decode

        Returns:
            Union[ContainerClient, FileSystemClient]
        """
        account_url, _, file_system, _ = BlobPathDecoder(path).get_with_url()
        return self._get_container_client(
            account_url=account_url,
            file_system=file_system)

    @abstractmethod
    def _get_container_client(
            self,
            account_url: str,
            file_system: str):
        """
        abstract method to be implemented
        :param account_url:
        :param file_system:
        :return:
        """
        raise NotImplementedError

    def ls(self, path: str, file_path: str):
        return self._ls(path=path, file_path=file_path)

    @abstractmethod
    def _ls(self, path: str, file_path: str):
        """
        abstract method to be implemented
        :param path:
        :param file_path:
        :return:
        """
        raise NotImplementedError

    def get(self, path: str, offset: int = None, length: int = None, **kwargs):
        """
        download data from Azure Blob or DataLake.

        Args:
            path:
            offset:
            length:
            **kwargs:

        Returns:

        """
        return self._get(path=path, offset=offset, length=length, **kwargs)

    @abstractmethod
    def _get(self, path: str, offset: int = None, length: int = None, **kwargs):
        """

        Args:
            path:
            offset:
            length:
            **kwargs:

        Returns:

        """
        raise NotImplementedError

    def put(self, path: str, data):
        return self._put(path=path, data=data)

    @abstractmethod
    def _put(self, path: str, data):
        """
        abstract method to be implemented
        :param path:
        :param data:
        :return:
        """
        raise NotImplementedError

    def create(self, path: str):
        return self._create(path=path)

    @abstractmethod
    def _create(self, path: str):
        raise NotImplementedError

    def append(self, path: str, data, offset: int):
        return self._append(path=path, data=data, offset=offset)

    @abstractmethod
    def _append(self, path: str, data, offset: int):
        raise NotImplementedError

    def info(self, path: str):
        return self._info(path=path)

    @abstractmethod
    def _info(self, path: str):
        """
        abstract method to be implemented
        :param path:
        :return:
        """
        raise NotImplementedError

    def rm(self, path: str):
        return self._rm(path=path)

    @abstractmethod
    def _rm(self, path: str):
        """
        abstract method to be implemented
        :param path:
        :return:
        """
        raise NotImplementedError



import functools
from inspect import signature
from typing import List, Dict

from azfs.error import AzfsInputError

from azure.cosmosdb.table.tableservice import TableService


class TableStorage:
    """
    A class for manipulating TableStorage in Storage Account
    The class provides simple methods below.

    * create
    * read
    * update
    * delete(not yet)

    The class is intended to be used as `delegation`, not `extend`.

    Args:
        account_name: name of the Storage Account
        account_key: key for the Storage Account
        database_name: name of the StorageTable database

    """

    def __init__(self, account_name: str, account_key: str, database_name: str):
        """

        """
        self.table_service = TableService(account_name=account_name, account_key=account_key)
        self.database_name = database_name

    def get(self, partition_key_value: str, filter_key_values: dict):
        """
        use PartitionKey as table_name

        Args:
            partition_key_value:
            filter_key_values:

        Returns:

        """
        filter_value_list = [
            f"PartitionKey eq '{partition_key_value}'"
        ]

        # その他の条件を付与する場合
        filter_value_list.extend(
            [f"{k} eq '{v}'" for k, v in filter_key_values.items()]
        )

        tasks = self.table_service.query_entities(table_name=self.database_name, filter=" and ".join(filter_value_list))
        return [task for task in tasks]

    def put(self, partition_key_value: str, data: dict):
        """
        put data.

        Args:
            partition_key_value:
            data:

        Returns:

        """
        insert_data = {'PartitionKey': partition_key_value}
        insert_data.update(data)
        self.table_service.insert_entity(table_name=self.database_name, entity=insert_data)
        return insert_data

    def update(self, partition_key_value: str, row_key: str, data: dict):
        """
        update data.

        Args:
            partition_key_value:
            row_key:
            data:

        Returns:

        """
        updated_data = {'PartitionKey': partition_key_value, 'RowKey': row_key}
        updated_data.update(data)
        self.table_service.update_entity(table_name=self.database_name, entity=updated_data)
        return updated_data


class TableStorageWrapper:
    """
    Wrapper for the TableStorage class.

    Args:
        account_name: name of the Storage Account
        account_key: key for the Storage Account
        database_name: name of the StorageTable database
        partition_key:
        row_key_name:

    Examples:
        >>> import json
        >>> from datetime import datetime
        >>> from pytz import timezone
        >>> tokyo = timezone('Asia/Tokyo')
        >>> cons = {
        ...     "account_name": "{storage_account_name}",
        ...     "account_key": "{credential}",
        ...     "database_name": "{database_name}"
        ... }
        # you can manipulate data through `simple_table_client`
        >>> simple_table_client = TableStorageWrapper(partition_key="simple_table", **cons)
        # store data according to the keyword-arguemnt you put
        # by default, `id_` is converted to `RowKey`, then `id_` is not stored
        >>> simple_table_client.put(id_="1", message="hello_world")
        ... {'PartitionKey': 'simple_table', 'message': 'hello_world', 'RowKey': '1'}
        # can get all data, simply call
        >>> simple_table_client.get()
        ... ...
        # or filter with specific value, like
        # `id_` is configured as `RowKey` by default
        >>> simple_table_client.get(id_="1")
        ... [
        ...     {
        ...         'PartitionKey': 'simple_table',
        ...         'RowKey': '1',
        ...         'Timestamp': datetime.datetime(2020, 10, 10, 3, 15, 57, 874427, tzinfo=tzutc()),
        ...         'message': 'hello_world',
        ...         'etag': 'W/"datetime\'2020-10-10T03%3A15%3A57.8744271Z\'"'
        ...     }
        ... ]
        # In addition, you can store data in different way
        >>> complex_client = TableStorageWrapper(partition_key="complex_table", **cons)
        >>> @complex_client.overwrite_pack_data_to_put()
        ... def modify_put_data(id_: str, message: str):
        ...     alt_message = json.dumps({datetime.now(tz=tokyo).isoformat(): message}, ensure_ascii=False)
        ...     return {"id_": id_, "message": alt_message}
        # you can store data in a different way
        >>> complex_client.put(id_="2", message="hello_world")
        ... {
        ...     'PartitionKey': 'complex_table',
        ...     'message': '{"2020-10-10T12:26:57.442718+09:00": "hello_world"}',
        ...     'RowKey': '2'
        ... }
        # you can also modify update function, with restriction example
        >>> @complex_client.overwrite_pack_data_to_update(allowed={"message": ["ERROR", "RUNNING", "SUCCESS"]})
        ... def modify_update_data(id_: str, message: str):
        ...     d = complex_client.get(id_=id_)
        ...     message_dict = json.loads(d[0]['message'])
        ...     if type(message_dict) is dict:
        ...         message_dict[datetime.now(tz=tokyo).isoformat()] = message
        ...     else:
        ...         message_dict = {datetime.now(tz=tokyo).isoformat(): message}
        ...
        ...     data = {
        ...         "id_": id_,
        ...         "message": json.dumps(message_dict, ensure_ascii=False)
        ...     }
        ...     return data
        >>> complex_client.update(id_="2", message="RUNNING")
        ... {
        ...     'PartitionKey': 'complex_table',
        ...     'RowKey': '2',
        ...     'message': '{"2020-10-10T12:26:57.442718+09:00": "hello_world", "2020-10-10T13:00:23.602943+09:00": "RUNNING"}'
        ... }
    """
    def __init__(
            self,
            account_name,
            account_key,
            database_name,
            partition_key: str,
            row_key_name: str = "id_"):
        self.st = TableStorage(account_name=account_name, account_key=account_key, database_name=database_name)
        self.partition_key = partition_key
        self.row_key_name = row_key_name

    @staticmethod
    def _check_argument(function: callable, arg_name: str) -> None:
        """
        Check whether the `arg_name` in the parameter of the `function`.

        Args:
            function: function to check
            arg_name: argument name

        Raises:
            ArgumentNameInvalidError: When `arg_name` is not found in the parameter of the `function`

        """
        # check arguments
        sig = signature(function)
        if arg_name not in sig.parameters:
            raise AzfsInputError(f"{arg_name} not in {function.__name__}")

    def _overwrite_pack_data(self, _to: str, allowed: Dict[str, List[str]] = None):
        """

        Args:
            _to:
            allowed:

        Returns:

        """
        def _wrapper(function: callable):
            # check if `row_key` argument exists
            self._check_argument(function=function, arg_name=self.row_key_name)

            def _actual_function(*args, **kwargs):
                # check if the value exists on the kwargs
                if allowed:
                    for k, v in allowed.items():
                        if kwargs[k] not in v:
                            raise AzfsInputError(f"keyword argument {kwargs[k]} is not allowed in {k}. {v} are allowed")
                return function(*args, **kwargs)

            # overwrite the attribute
            setattr(self, _to, _actual_function)

            return function
        return _wrapper

    # =======
    # get
    # =======

    def get(self, **kwargs) -> List[dict]:
        """
        get data.

        Args:
            **kwargs:

        Returns:

        """
        if self.row_key_name in kwargs:
            kwargs['RowKey'] = kwargs.pop(self.row_key_name)
        return self.st.get(partition_key_value=self.partition_key, filter_key_values=kwargs)

    # =======
    # put
    # =======

    @staticmethod
    def pack_data_to_put(**kwargs):
        """

        Args:
            **kwargs:

        Returns:

        """
        return kwargs

    def put(self, **kwargs):
        """

        Args:
            **kwargs:

        Returns:

        """
        _data = self.pack_data_to_put(**kwargs)
        if self.row_key_name in _data:
            _data['RowKey'] = _data.pop(self.row_key_name)
        return self.st.put(partition_key_value=self.partition_key, data=_data)

    overwrite_pack_data_to_put = functools.partialmethod(_overwrite_pack_data, _to="pack_data_to_put")

    # =======
    # update
    # =======

    @staticmethod
    def pack_data_to_update(**kwargs):
        """

        Args:
            **kwargs:

        Returns:

        """
        return kwargs

    def update(self, **kwargs):
        """

        Args:
            **kwargs:

        Returns:

        """
        _data = self.pack_data_to_update(**kwargs)
        row_key = _data.pop(self.row_key_name)
        return self.st.update(partition_key_value=self.partition_key, row_key=row_key, data=_data)

    overwrite_pack_data_to_update = functools.partialmethod(_overwrite_pack_data, _to="pack_data_to_update")



import logging
from typing import Union, Optional

from fsspec import AbstractFileSystem
from fsspec.spec import AbstractBufferedFile
from azure.identity import DefaultAzureCredential

from .utils import (
    BlobPathDecoder,
    ls_filter
)
from .clients import AzfsClient


logger = logging.getLogger('azfs')
__all__ = ["AzFileSystem", "AzFile"]


class AzFileSystem(AbstractFileSystem):
    default_block_size = 5 * 2 ** 20

    def __init__(
            self,
            credential: Optional[Union[str, DefaultAzureCredential]] = None,
            connection_string: str = None,
            *args,
            **storage_options):
        super().__init__(*args, **storage_options)
        if credential is None and connection_string is None:
            credential = DefaultAzureCredential()
        self.az_client = AzfsClient(credential=credential, connection_string=connection_string)

    def _open(
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=True,
        cache_options=None,
        **kwargs
    ):
        """
        (Override method)
        Return raw bytes-mode file-like from the file-system

        """
        if block_size is None:
            block_size = self.default_block_size

        return AzFile(
            self,
            path,
            mode,
            block_size,
            autocommit,
            cache_options=cache_options,
            **kwargs
        )

    def info(self, path, **kwargs):
        """
        (Override method)

        Args:
            path:
            **kwargs:

        Returns:

        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        # get info from blob or data-lake storage
        data = self.az_client.get_client(account_kind=account_kind).info(path=path)

        # extract below to determine file or directory
        content_settings = data.get("content_settings", {})
        metadata = data.get("metadata", {})

        data_type = ""
        if "hdi_isfolder" in metadata:
            # only data-lake storage has `hdi_isfolder`
            data_type = "directory"
        elif content_settings.get("content_type") is not None:
            # blob and data-lake storage have `content_settings`,
            # and its value of the `content_type` must not be None
            data_type = "file"
        return {
            "name": data.get("name", ""),
            "size": data.get("size", ""),
            "creation_time": data.get("creation_time", ""),
            "last_modified": data.get("last_modified", ""),
            "etag": data.get("etag", ""),
            "content_type": content_settings.get("content_type", ""),
            "type": data_type
        }

    def ls(self, path, detail=True, attach_prefix=False, **kwargs):
        """
        (Inherited method)

        Args:
            path:
            detail:
            attach_prefix:
            **kwargs:

        Returns:

        """
        _, account_kind, _, file_path = BlobPathDecoder(path).get_with_url()
        file_list = self.az_client.get_client(account_kind=account_kind).ls(path=path, file_path=file_path)
        if account_kind in ["dfs", "blob"]:
            file_name_list = ls_filter(file_path_list=file_list, file_path=file_path)
            if attach_prefix:
                path = path if path.endswith("/") else f"{path}/"
                file_full_path_list = [f"{path}{f}" for f in file_name_list]
                return file_full_path_list
            else:
                return file_name_list

    def copy(self, path1, path2, **kwargs):
        """
        (Inherited method)


        Args:
            path1:
            path2:
            **kwargs:

        Returns:

        """
        pass

    def _rm(self, path):
        """
        (Inherited method)

        Args:
            path:

        Returns:

        """
        pass

    def created(self, path):
        """
        (Inherited method)


        Args:
            path:

        Returns:

        """
        pass

    def modified(self, path):
        """
        (Inherited method)

        Args:
            path:

        Returns:

        """
        pass


class AzFile(AbstractBufferedFile):
    """

    Args:
        fs: AzFileSystem
        path: path to read file
        mode: value candidates are ["ab", "rb", "wb"], same as build-in function `open()`
        block_size: Buffer size for reading or writing, 'default' for class default
        autocommit:
        cache_type:
        cache_option:
    """

    part_max = 5 * 2 ** 30

    def __init__(
            self,
            fs: AzFileSystem,
            path: str,
            mode="rb",
            block_size="default",
            autocommit=True,
            cache_type="readahead",
            cache_options=None,
            **kwargs):
        # path, mode, start, end, block_size, and etc, are initialized
        # if mode is "rb", prepare to read data in azure
        # otherwise ("ab" or "wb"), to write data to azure
        super().__init__(fs, path, mode, block_size, autocommit, cache_type, cache_options, **kwargs)
        # check block_size from AbstractBufferedFile
        if self.writable():
            if block_size < 5 * 2 ** 20:
                raise ValueError('Block size must be >=5MB')
        # when not using autocommit we want to have transactional state to manage
        self.append_block = False

        # reference by s3fs
        if 'a' in mode and fs.exists(path):
            loc = fs.info(path)['size']
            if loc < 5 * 2 ** 20:
                # existing file too small for multi-upload: download
                self.write(self.fs.cat(self.path))
            else:
                self.append_block = True
            self.loc = loc

    def _upload_chunk(self, final=False):
        """
        (Override method)
        Write one part of a multi-block file upload

        Args:
            final: This is the last block, so should complete file, if self.autocommit is True.

        Returns:

        """
        logger.debug("upload chunk")
        # decode azure path
        _, account_kind, _, file_path = BlobPathDecoder(self.path).get_with_url()

        # may not yet have been initialized, may need to call _initialize_upload
        if self.autocommit and not self.append_block and final and self.tell() < self.blocksize:
            # only happens when closing small file, use on-shot PUT
            data1 = False
        else:
            self.buffer.seek(0)
            (data0, data1) = (None, self.buffer.read(self.blocksize))

        (offset0, offset1) = (0, 0)
        while data1:
            (offset0, offset1) = (offset1, self.buffer.tell())
            (data0, data1) = (data1, self.buffer.read(self.blocksize))
            data1_size = len(data1)

            logger.debug(f"data0_size: {len(data0)}")

            if 0 < data1_size < self.blocksize:
                remainder = data0 + data1
                remainder_size = self.blocksize + data1_size

                if remainder_size <= self.part_max:
                    (data0, data1) = (remainder, None)
                else:
                    partition = remainder_size // 2
                    (data0, data1) = (remainder[:partition], remainder[partition:])

            # part = len(self.parts) + 1
            logger.debug(f"offset0: {offset0}")
            # logger.debug(f"self.buffer.tell(): {self.buffer.tell()}")

            try:
                _ = self.fs.az_client.get_client(account_kind=account_kind) \
                        .append(path=self.path, data=data0, offset=offset0)
            except Exception as e:
                raise IOError from e

        if self.autocommit and final:
            self.commit()
        return not final

    def commit(self):
        logger.debug("commit")
        # decode azure path
        _, account_kind, _, file_path = BlobPathDecoder(self.path).get_with_url()

        # read data
        if self.autocommit and not self.append_block and self.tell() < self.blocksize:
            # only happens when closing small file, use on-shot PUT
            self.buffer.seek(0)
            data = self.buffer.read(self.blocksize)
            _ = self.fs.az_client.get_client(account_kind=account_kind).put(path=self.path, data=data)

    def _initiate_upload(self):
        """
        (Override method)
        Create remote file/upload

        Returns:

        """
        logger.debug("initiate upload")
        if self.autocommit and not self.append_block and self.tell() < self.blocksize:
            # only happens when closing small file, use on-shot PUT
            return
        _, account_kind, _, file_path = BlobPathDecoder(self.path).get_with_url()
        self.fs.az_client.get_client(account_kind=account_kind).create(path=self.path)

    def _fetch_range(self, start, end):
        """
        (Inherited method)

        Args:
            start:
            end:

        Returns:

        """
        _, account_kind, _, file_path = BlobPathDecoder(self.path).get_with_url()
        return self.fs.az_client.get_client(
            account_kind=account_kind).get(path=self.path, offset=start, length=end-start)



import bz2
import gzip
import io
import json
import lzma
import os
import pickle
import sys
import pandas as pd
import pytest
import azfs

# テスト対象のファイルへのパスを通している
# pytestの設定
PARENT_PATH = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
SOURCE_PATH = PARENT_PATH.rsplit('/', 1)[0]

sys.path.append(f"{SOURCE_PATH}")


@pytest.fixture()
def _get_csv(mocker):
    """
    original data is
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")
    :param mocker:
    :return:
    """
    return_value = b'name,age\nalice,10\nbob,10\n'
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _get_csv_gz(mocker):
    """
    original data is
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")

    Args:
        mocker:

    Returns:

    """
    data = b'name,age\nalice,10\nbob,10\n'
    return_value = gzip.compress(data)
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _get_table(mocker):
    """
    original data is
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")
    :param mocker:
    :return:
    """
    return_value = b'\tname\tage\n1\talice\t10\n2\tbob\t10\n'
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _get_pickle(mocker):
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")
    p = pickle.dumps(df)
    return_value = io.BytesIO(p)
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _get_pickle_gzip(mocker):
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")
    p = pickle.dumps(df)
    return_value = io.BytesIO(gzip.compress(p))
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _get_pickle_bz2(mocker):
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")
    p = pickle.dumps(df)
    return_value = io.BytesIO(bz2.compress(p))
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _get_pickle_xz(mocker):
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")
    p = pickle.dumps(df)
    return_value = io.BytesIO(lzma.compress(p))
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _get_json(mocker):
    """
    :param mocker:
    :return:
    """
    return_value = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    func_mock = mocker.MagicMock()
    func_mock.return_value = json.dumps(return_value)
    yield func_mock


@pytest.fixture()
def _put(mocker):
    """
    :param mocker:
    :return:
    """
    return_value = True
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _ls(mocker):
    """
    :param mocker:
    :return:
    """
    return_value = ["test1.csv", "test2.csv", "dir/"]
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _ls_for_glob(mocker):
    """
    :param mocker:
    :return:
    """
    return_value = [
        "root_folder/test1.csv",
        "root_folder/test2.csv",
        "root_folder/test1.json",
        "root_folder/dir1/test1.csv",
        "root_folder/dir1/test2.csv",
        "root_folder/dir1/test1.json",
        "root_folder/dir2/test1.csv",
        "root_folder/dir2/test2.csv",
        "root_folder/dir2/test1.json",
    ]
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def _rm(mocker):
    """
    :param mocker:
    :return:
    """
    return_value = True
    func_mock = mocker.MagicMock()
    func_mock.return_value = return_value
    yield func_mock


@pytest.fixture()
def var_json() -> pd.DataFrame:
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    yield data


@pytest.fixture()
def var_df() -> pd.DataFrame:
    data = {"1": {"name": "alice", "age": "10"}, "2": {"name": "bob", "age": "10"}}
    df = pd.DataFrame.from_dict(data, orient="index")
    yield df


@pytest.fixture()
def var_azc() -> azfs.AzFileClient:
    azc = azfs.AzFileClient()
    yield azc



import setuptools
import os


def get_version(init_file_path):
    version_line = list(
        filter(lambda l: l.startswith('VERSION'), open(init_file_path))
    )[0]

    # eval is required to convert from string to tuple,
    # because VERSION defined in __init__.py is tuple
    version_tuple = eval(version_line.split('=')[-1])

    # join with dot
    return ".".join(map(str, version_tuple))


# get __version__ from __init__.py
init = os.path.join(
    os.path.dirname(__file__), 'azfs', '__init__.py'
)
VERSION = get_version(init_file_path=init)

with open("README.md", "r") as fh:
    long_description = fh.read()

PROJECT_URLS = {
    "Bug Tracker": "https://github.com/gsy0911/azfs/issues",
    "Documentation": "https://azfs.readthedocs.io/en/latest/?badge=latest",
    "Source Code": "https://github.com/gsy0911/azfs",
}


setuptools.setup(
    name="azfs",
    version=VERSION,
    author="gsy0911",
    author_email="yoshiki0911@gmail.com",
    description="AzFS is to provide convenient Python read/write functions for Azure Storage Account.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gsy0911/azfs",
    project_urls=PROJECT_URLS,
    packages=setuptools.find_packages(),
    install_requires=[
        "pandas",
        "azure-cosmosdb-table",
        "azure-identity>=1.3.1",
        "azure-storage-blob>=12.3.0",
        "azure-storage-file-datalake>=12.0.0",
        "azure-storage-queue>=12.1.1",
        "fsspec"
    ],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License"
    ],
    python_requires='>=3.6',
    keywords=["Azure", "StorageAccount", "Blob", "DataLake", "Queue"]
)
